XIncludeFile "disconnect.pbi"
XIncludeFile "http.pbi"

; For our image in the dialog
UsePNGImageDecoder()

; Files to download
Structure DownloadFiles
	url.s
	destination.s
	overwrite.i
	status.i
	md5.s
EndStructure

; Possible states
Enumeration
	#DOWNLOAD_STATUS_PENDING 				; Wait until previous downloads finished
	#DOWNLOAD_STATUS_INPROGRESS 		; Currently in progress
	#DOWNLOAD_STATUS_DONE 					; Successful
	#DOWNLOAD_STATUS_FAILED 				; Failed (unable to connect, aborted etc)
	#DOWNLOAD_STATUS_CHECKSUMERROR	; Checksum error
	#DOWNLOAD_STATUS_EXISTS 				; File exists already
EndEnumeration

; Size of the internal download buffer
#DOWNLOAD_BUFFER = 102400

; Download dialog structure
Structure DownloadDialog
	
	; Configuration
	hideWhenFinished.i	
	
	; List of files to download
	List files.DownloadFiles()
	started.i
	
	; Last progress status
	currentProgress.d
	speed.d
	speed_text.s
	eta.s
	
	; Current opened file / allocated memory
	*http
	file.i
	md5.i
	*memory
	
	; Start time and start bytes
	estimateStartTime.i
	estimateStartBytes.i
	
	; GUI stuff
	window.i
	image.i

	progressbar_overall.i
	progressbar_current.i
	
	label_currentfile.i
	label_url.i
	label_destination.i
	label_eta.i
	label_speed.i
EndStructure

#DOWNLOAD_WINDOW_WIDTH = 500

CompilerIf #PB_Compiler_OS = #PB_OS_Linux
	ImportC ""
		gtk_progress_bar_set_text(*pbar, text.p-ascii)
	EndImport
CompilerEndIf

DataSection
	DownloadDialogImage:
		IncludeBinary "dl.png"

EndDataSection

Structure DownloadDialog_Ptr
	*struct.DownloadDialog
EndStructure

Global NewList AllocatedDownloadDialogs.DownloadDialog_Ptr()

Procedure DownloadDialog_Show(*dialog.DownloadDialog)
  If *dialog = 0
    Debug "No valid DownloadDialog object given"
    ProcedureReturn #False
  EndIf
	
	; Window already exists?
	If *dialog\window
		ProcedureReturn #True
	EndIf
	
	; Open window
	*dialog\window = OpenWindow(#PB_Any, 0, 0, #DOWNLOAD_WINDOW_WIDTH, 330, "Downloader", #PB_Window_SystemMenu | #PB_Window_ScreenCentered)
	
	; Failed to create new window
	If *dialog\window = 0
		ProcedureReturn #False
	EndIf

	Protected Row.i = 5
	
	With *dialog

		Protected Image =  CatchImage(#PB_Any, ?DownloadDialogImage)
		If Image
			Protected PosX = (#DOWNLOAD_WINDOW_WIDTH) / 2 - 64
			\image								=	ImageGadget(			#PB_Any,	PosX, 	Row, 129, 110, ImageID(Image))
			FreeImage(Image)
		EndIf

		Row + 120

															TextGadget(				#PB_Any, 		5, 		Row, 100, 25, "Overall: ")
		\progressbar_overall		=	ProgressBarGadget(#PB_Any, 	110, 		Row, #DOWNLOAD_WINDOW_WIDTH - 120, 25, 0, 100)
		Row + 30
	
															TextGadget(				#PB_Any, 		5, 		Row, 100 - 10, 25, "File: ")
		\label_currentfile			=	TextGadget(				#PB_Any, 	110, 		Row, #DOWNLOAD_WINDOW_WIDTH - 10, 25, "/")
		Row + 30
		
															TextGadget(				#PB_Any, 	5, 			Row, 100, 25, "Current: ")
		\progressbar_current		=	ProgressBarGadget(#PB_Any, 	110, 		Row, #DOWNLOAD_WINDOW_WIDTH - 120, 25, 0, 100)
		Row + 30			
		
															TextGadget(				#PB_Any, 	5, 			Row, 100, 25, "Source: ")
		\label_url							=	TextGadget(				#PB_Any, 	110, 		Row, #DOWNLOAD_WINDOW_WIDTH - 10, 25, "")
		Row + 30

															TextGadget(				#PB_Any, 	5, 			Row, 100, 25, "Destination: ")
		\label_destination			=	TextGadget(				#PB_Any, 	110, 		Row, #DOWNLOAD_WINDOW_WIDTH - 10, 25, "")
		Row + 30
		
															TextGadget(				#PB_Any, 	5, 			Row, 100, 25, "ETA: ")
		\label_eta							=	TextGadget(				#PB_Any, 	110, 		Row, #DOWNLOAD_WINDOW_WIDTH - 10, 25, "")
		Row + 30			

															TextGadget(				#PB_Any, 	5, 			Row, 100, 25, "Speed: ")
		\label_speed						=	TextGadget(				#PB_Any, 	110, 		Row, #DOWNLOAD_WINDOW_WIDTH - 10, 25, "")
		Row + 30
							
	EndWith
		
	ProcedureReturn #True
EndProcedure

Procedure DownloadDialog_Hide(*dialog.DownloadDialog)
  If *dialog = 0
    Debug "No valid DownloadDialog object given"
    ProcedureReturn #False
  EndIf
	
	; Window already exists?
	If *dialog\window
		CloseWindow(*dialog\window)
		*dialog\window = 0
	EndIf

	ProcedureReturn #True
EndProcedure

Procedure DownloadDialog_Create(show.i = #True, hideWhenFinished.i = #True)
	Protected *dialog.DownloadDialog = AllocateMemory(SizeOf(DownloadDialog))
	
	If *dialog = 0
		Debug "Unable to allocate memory for DownloadDialog"
		ProcedureReturn 0
	EndIf
	
	; Remember the pointer in order to handle window events
	If AddElement(AllocatedDownloadDialogs())
		AllocatedDownloadDialogs()\struct = *dialog
	EndIf
	
	InitializeStructure(*dialog, DownloadDialog)
	
	*dialog\hideWhenFinished 	= hideWhenFinished
	*dialog\started 					= #False
	
	If show
		DownloadDialog_Show(*dialog)
	EndIf
	
	ProcedureReturn *dialog
EndProcedure

Procedure DownloadDialog_Free(*dialog.DownloadDialog)
  If *dialog = 0
    Debug "No valid DownloadDialog object given"
    ProcedureReturn #False
  EndIf
  
  If *dialog\http
  	HTTPQuery_Free(*dialog\http)
 		*dialog\http = 0
  EndIf
  
  If *dialog\file
  	CloseFile(*dialog\file)
  	*dialog\file = 0
 	EndIf
 	
 	If *dialog\memory
 		FreeMemory(*dialog\memory)
 		*dialog\memory = 0
 	EndIf
 	
 	If *dialog\window
 		CloseWindow(*dialog\window)
 		*dialog\window = 0
 	EndIf
  
  ; Remove from list of allocated download dialogs
  ForEach AllocatedDownloadDialogs()
  	If AllocatedDownloadDialogs()\struct = *dialog
  		DeleteElement(AllocatedDownloadDialogs())
  		Break
  	EndIf
  Next
  
  ClearStructure(*dialog, DownloadDialog)
  FreeMemory(*dialog)
EndProcedure

Procedure DownloadDialog_AddFile(*dialog.DownloadDialog, url.s, destinationPath.s, overwrite.i = #True, md5.s = "")
  If *dialog = 0
    Debug "No valid DownloadDialog object given"
    ProcedureReturn #False
  EndIf
  
	If Not *dialog\started
		If AddElement(*dialog\files())
			*dialog\files()\url 				= url
			*dialog\files()\destination = destinationPath
			*dialog\files()\overwrite		= overwrite
			*dialog\files()\status			= #DOWNLOAD_STATUS_PENDING
			*dialog\files()\md5 				= md5
			
			ProcedureReturn #True
		EndIf
	EndIf
	
	ProcedureReturn #False
EndProcedure

Procedure DownloadDialog_ClearFiles(*dialog.DownloadDialog)
  If *dialog = 0
    Debug "No valid DownloadDialog object given"
    ProcedureReturn #False
  EndIf
  
  If Not *dialog\started
  	ClearList(*dialog\files())
  	ProcedureReturn #True
  EndIf
  
  ProcedureReturn #False
EndProcedure

Procedure DownloadDialog_UpdateWindow(*dialog.DownloadDialog)
  If *dialog = 0
    Debug "No valid DownloadDialog object given"
    ProcedureReturn #False
  EndIf

	; Nothing to update?
	If *dialog\window = 0
		ProcedureReturn #False
	EndIf	
	
	; Window already closed?
	;If Not IsWindow(*dialog\window)
	;	*dialog\window = 0
	;	ProcedureReturn #False
	;EndIf
	
	Protected fileCount.i 		= ListSize(*dialog\files())
	Protected currentFile.i 	= ListIndex(*dialog\files())
	
	Protected currentFile_text.s = Str(currentFile + 1) + "/" + Str(fileCount)
	
	If GetGadgetText(*dialog\label_currentfile) <> currentFile_text
	  SetGadgetText(*dialog\label_currentfile, currentFile_text)
	EndIf
	
	If *dialog\started And currentFile >= 0
		SetGadgetState(*dialog\progressbar_overall, ((currentFile + *dialog\currentProgress) * 100) / fileCount)
		SetGadgetState(*dialog\progressbar_current, *dialog\currentProgress * 100)
		
		CompilerIf #PB_Compiler_OS = #PB_OS_Linux
			gtk_progress_bar_set_text(GadgetID(*dialog\progressbar_overall), Str(((currentFile + *dialog\currentProgress) * 100) / fileCount) + "%")
			gtk_progress_bar_set_text(GadgetID(*dialog\progressbar_current), Str(*dialog\currentProgress * 100) + "%")
		CompilerEndIf
		
		If GetGadgetText(*dialog\label_url) <> *dialog\files()\url
		  SetGadgetText(*dialog\label_url, 					*dialog\files()\url)
		EndIf
		
		If GetGadgetText(*dialog\label_destination) <> *dialog\files()\destination
		  SetGadgetText(*dialog\label_destination, 	*dialog\files()\destination)
		EndIf
		
		If GetGadgetText(*dialog\label_eta) <> *dialog\eta
		  SetGadgetText(*dialog\label_eta, 					*dialog\eta)
		EndIf
		
		If GetGadgetText(*dialog\label_speed) <> *dialog\speed_text
		  SetGadgetText(*dialog\label_speed, 				*dialog\speed_text)		
		EndIf
		
	Else
		SetGadgetState(*dialog\progressbar_overall, 100)
		SetGadgetState(*dialog\progressbar_current, 0)

		CompilerIf #PB_Compiler_OS = #PB_OS_Linux
			gtk_progress_bar_set_text(GadgetID(*dialog\progressbar_overall), "")
			gtk_progress_bar_set_text(GadgetID(*dialog\progressbar_current), "")
		CompilerEndIf
		
		If GetGadgetText(*dialog\label_url) <> ""
		  SetGadgetText(*dialog\label_url, 					"")
		EndIf
		
		If GetGadgetText(*dialog\label_destination) <> ""
		  SetGadgetText(*dialog\label_destination, 	"")
		EndIf
		
		If GetGadgetText(*dialog\label_eta) <> ""
		  SetGadgetText(*dialog\label_eta, 					"")	
		EndIf
		
		If GetGadgetText(*dialog\label_speed) <> ""
		  SetGadgetText(*dialog\label_speed, 				"")
		EndIf
	EndIf
	
	ProcedureReturn #True
EndProcedure

Procedure DownloadDialog__Connect_(*dialog.DownloadDialog)
  If *dialog = 0
    Debug "No valid DownloadDialog object given"
    ProcedureReturn #False
  EndIf
  
	;Not started yet
	If Not *dialog\started
		ProcedureReturn #False
	EndIf
	
	; Close file if it is already opened?
	If *dialog\file
		CloseFile(*dialog\file)
		*dialog\file = 0
	EndIf
	
	If *dialog\md5
	  FinishFingerprint(*dialog\md5)
	  *dialog\md5 = 0
	EndIf
	
	If *dialog\memory = 0
		*dialog\memory = AllocateMemory(#DOWNLOAD_BUFFER)
	EndIf
	
	If *dialog\memory = 0
		Debug "Unable to allocate download buffer"
		ProcedureReturn #False
	EndIf

	; Initialize progress variables
	*dialog\currentProgress = 0.0
	*dialog\speed           = 0.0
	*dialog\speed_text			= ""
	*dialog\eta             = ""

	While NextElement(*dialog\files())
	
		; Update the window
		DownloadDialog_UpdateWindow(*dialog)
		
		; Is this file alreday processed?
		If *dialog\files()\status <> #DOWNLOAD_STATUS_PENDING
			Continue
		EndIf
		
		; If the file already exists, skip over it
		If FileSize(*dialog\files()\destination) >= 0 And Not *dialog\files()\overwrite
			*dialog\files()\status = #DOWNLOAD_STATUS_EXISTS
			Continue
		EndIf
		
		; Can we open the file?
		*dialog\file = CreateFile(#PB_Any, *dialog\files()\destination)
		If Not *dialog\file
			*dialog\files()\status = #DOWNLOAD_STATUS_FAILED
			Continue
		EndIf
		
		*dialog\http = HTTPQuery_New(*dialog\files()\url)
		If *dialog\http
			
			; Try to open the connection
			If HTTPQuery_Open(*dialog\http)
				Protected statuscode.i = HTTPQuery_GetStatuscode(*dialog\http)
				
				; Check for valid statuscode
				If statuscode >= 200 And statuscode < 300
					
					; Download is in progress
					*dialog\files()\status = #DOWNLOAD_STATUS_INPROGRESS
					
					; Allocate a new fingerprint... dont care if it fails
					If *dialog\files()\md5 <> ""
						*dialog\md5 = ExamineMD5Fingerprint(#PB_Any)
					EndIf
				
					; Start time and start bytes
					*dialog\estimateStartTime 	= ElapsedMilliseconds()
					*dialog\estimateStartBytes 	= 0
				
					ProcedureReturn #True
				EndIf
				
				HTTPQuery_Close(*dialog\http)
			EndIf
		
			HTTPQuery_Free(*dialog\http)
			*dialog\http = 0
		EndIf
		
		; Close the file
		CloseFile(*dialog\file)	
		*dialog\file = 0
		
		*dialog\files()\status = #DOWNLOAD_STATUS_FAILED
		
	Wend
	
	; Not started anymore
	*dialog\started = #False
	
	If *dialog\hideWhenFinished
		DownloadDialog_Hide(*dialog)
	EndIf
	
	; Start with an empty list again?
	;ClearList(*dialog\files())
	
	ProcedureReturn #True
EndProcedure

Procedure DownloadDialog_Start(*dialog.DownloadDialog)
  If *dialog = 0
    Debug "No valid DownloadDialog object given"
    ProcedureReturn #False
  EndIf

	; Already started
	If *dialog\started
		ProcedureReturn #True
	EndIf
	
	; Do we have something to download?
	If ListSize(*dialog\files()) <= 0
		ProcedureReturn #True
	EndIf	
	
	; Begin download
	*dialog\started = #True
	ResetList(*dialog\files())
	
	; Try to connect and start with the first file that works
	ProcedureReturn DownloadDialog__Connect_(*dialog)
EndProcedure

Procedure DownloadDialog__Events_(*dialog.DownloadDialog)
	If *dialog = 0
    Debug "No valid DownloadDialog object given"
    ProcedureReturn #False
  EndIf
  
	;Not started yet or variables not initialized correctly
	If Not *dialog\started Or *dialog\http = 0 Or *dialog\file = 0
		ProcedureReturn #False
	EndIf
	
	Protected changed.i = #False
	
	; No EOF yet	
	If Not HTTPQuery_EOF(*dialog\http)
		
		Repeat
			; Do nonblocking read
			Protected temp.i = HTTPQuery_Read(*dialog\http, *dialog\memory, #DOWNLOAD_BUFFER, #True)
			If temp > 0
				WriteData(*dialog\file, *dialog\memory, temp)
				
				; Update md5 sum
				If *dialog\md5
					NextFingerprint(*dialog\md5, *dialog\memory, temp)
				EndIf
				
        changed = #True
			EndIf
		Until temp = 0 ; Wait for more data
		
		; Update progress
		*dialog\currentProgress = HTTPQuery_GetProgress(*dialog\http)
	EndIf
	
	If changed
  
  	; Update speed & eta
  	Protected totalContentLength.i = HTTPQuery_GetTotalContentLength(*dialog\http)
  	Protected totalBytesReceived.i = HTTPQuery_GetTotalBytesReceived(*dialog\http)
  	Protected currentTime.i        = ElapsedMilliseconds()
  	
  	; Wait before first estimation
  	If currentTime > *dialog\estimateStartTime + 500
  		*dialog\speed = (totalBytesReceived - *dialog\estimateStartBytes) * 1000.0 / (currentTime - *dialog\estimateStartTime)
  	
  		If *dialog\speed < 1024
  			*dialog\speed_text = StrD(*dialog\speed / 1024, 2) + " kB/s"
  		Else
  			*dialog\speed_text = StrD(*dialog\speed / (1024 * 1024), 2) + " MB/s" 
  		EndIf
  		
  	EndIf
  	   
  	If currentTime > *dialog\estimateStartTime + 3000
  		*dialog\estimateStartTime 	= (*dialog\estimateStartTime + currentTime) / 2
  		*dialog\estimateStartBytes 	= (*dialog\estimateStartBytes + totalbytesReceived) / 2
  	EndIf
  	   
  	If totalContentLength = -1
  		*dialog\eta = "Total size unknown"
  	
  	ElseIf *dialog\speed > 0
  		Protected time.i = Round( (totalContentLength - totalBytesReceived) / *dialog\speed, #PB_Round_Up) ; Times in seconds
  		
  		If time < 60
  			*dialog\eta = Str(time) + " seconds remaining"
  		ElseIf time < 60*60
  			*dialog\eta = Str(time / 60) + " minutes remaining" ;+ Str(time % 60) + " seconds remaining"
  		Else
  			*dialog\eta = Str(time / 3600) + " hours, " + Str( (time % 3600) / 60 ) + " minutes remaining"
  		EndIf
  		
  	Else
  		*dialog\eta = "Please wait"
  	
  	EndIf
  	
  EndIf
  
	; EOF found
	If HTTPQuery_EOF(*dialog\http)
	
		; Everything okay
    If HTTPQuery_GetStatuscode(*dialog\http) >= 0
			*dialog\files()\status = #DOWNLOAD_STATUS_DONE
			
		Else
			*dialog\files()\status = #DOWNLOAD_STATUS_FAILED

		EndIf
		
		If *dialog\md5
		  Protected md5.s = FinishFingerprint(*dialog\md5)
		  *dialog\md5 = 0
		  
		  If *dialog\files()\md5 <> "" And LCase(*dialog\files()\md5) <> LCase(md5)
		  	*dialog\files()\status = #DOWNLOAD_STATUS_CHECKSUMERROR
		  EndIf
		  
		EndIf
		
		; Free the http object
		HTTPQuery_Free(*dialog\http)
		*dialog\http = 0
		
		; Close the file
		CloseFile(*dialog\file)
		*dialog\file = 0
		
		; Continue with the next file
		DownloadDialog__Connect_(*dialog)
		
	EndIf
	
	If changed
  	
  	; Update status
  	DownloadDialog_UpdateWindow(*dialog)
  	
  EndIf
  	
	If Not *dialog\started
		ProcedureReturn #True
	Else
		ProcedureReturn #False
	EndIf
EndProcedure

Procedure DownloadDialog_Status(*dialog.DownloadDialog)
  If *dialog = 0
    Debug "No valid DownloadDialog object given"
    ProcedureReturn #DOWNLOAD_STATUS_DONE
  EndIf
  
  If *dialog\started
  	ProcedureReturn #DOWNLOAD_STATUS_INPROGRESS
  EndIf
  
  Protected files_pending.i 	= #False
  Protected files_failed.i 	= #False
  
  ForEach *dialog\files()
  	If *dialog\files()\status = #DOWNLOAD_STATUS_PENDING Or *dialog\files()\status = #DOWNLOAD_STATUS_INPROGRESS
  		files_pending = #True
  	ElseIf *dialog\files()\status <> #DOWNLOAD_STATUS_DONE
  		files_failed  = #True
    EndIf
  Next
  
  If files_pending
  	ProcedureReturn #DOWNLOAD_STATUS_PENDING
  EndIf
  
  If files_failed
  	ProcedureReturn #DOWNLOAD_STATUS_FAILED
 	EndIf
 	
 	ProcedureReturn #DOWNLOAD_STATUS_DONE
EndProcedure

Procedure DownloadDialog_RetryFailed(*dialog.DownloadDialog)
  If *dialog = 0
    Debug "No valid DownloadDialog object given"
    ProcedureReturn #False
  EndIf
  
  If *dialog\started
  	ProcedureReturn #False
  EndIf
  
  Protected files_restarted.i = #False
  
  ForEach *dialog\files()
  	If *dialog\files()\status = #DOWNLOAD_STATUS_PENDING
  		Continue
  		
  	ElseIf *dialog\files()\status <> #DOWNLOAD_STATUS_DONE
  		*dialog\files()\status = #DOWNLOAD_STATUS_PENDING
  		files_restarted = #True	
  	
  	Else ; Done -> delete
  		DeleteElement(*dialog\files())
  	EndIf  
  Next
  
 	ProcedureReturn files_restarted
EndProcedure

Procedure DownloadDialog_GetFileStatus(*dialog.DownloadDialog, destinationPath.s)
  If *dialog = 0
    Debug "No valid DownloadDialog object given"
    ProcedureReturn -1
  EndIf
  
  If *dialog\started
  	ProcedureReturn -1
  EndIf
  
  ForEach *dialog\files()
  	If *dialog\files()\destination = destinationPath
  		ProcedureReturn *dialog\files()\status
  	EndIf  
  Next
  
  ProcedureReturn -1
EndProcedure

Procedure.s DownloadDialog_NextFailedFile(*dialog.DownloadDialog, *status.INTEGER = 0)
  If *dialog = 0
    Debug "No valid DownloadDialog object given"
    ProcedureReturn ""
  EndIf
  
  If *dialog\started
  	ProcedureReturn ""
  EndIf
	
	ForEach *dialog\files()
		If *dialog\files()\status = #DOWNLOAD_STATUS_PENDING
  		Continue
  		
  	ElseIf *dialog\files()\status <> #DOWNLOAD_STATUS_DONE
  		
  		; Save the status if the user wants to know the exact failure reason
  		If *status
  			*status\i = *dialog\files()\status
  		EndIf
  		
  		Protected destinationPath.s = *dialog\files()\destination
  		  	
  		DeleteElement(*dialog\files())
  		
  		ProcedureReturn destinationPath
  	EndIf  
 	Next
 	
 	ProcedureReturn ""
EndProcedure

Procedure.i Hook_WindowEvent( *forceReturn.INTEGER = 0 )

	If ListSize(AllocatedDownloadDialogs()) = 0
		ProcedureReturn WindowEvent()
	EndIf
	
  ; Event handling first
  ForEach AllocatedDownloadDialogs()
		If DownloadDialog__Events_(AllocatedDownloadDialogs()\struct)
			If *forceReturn <> 0
				*forceReturn\i = #True
			EndIf
		EndIf
 	Next
 	
 	; Parse events (also events caused by the status update)
	Repeat
	  
	  ; On windows the debugger complains if no window is opened ...
	  DisableDebugger
  	Protected Event.i 	= WindowEvent()
  	EnableDebugger
  	
  	If Event = 0
  	  Break
  	EndIf
  	
  	Protected Window.i = EventWindow()
  	
  	ForEach AllocatedDownloadDialogs()
  		If Event <> 0 And Window <> 0 And AllocatedDownloadDialogs()\struct\window = Window
  
  			If Event = #PB_Event_CloseWindow
  				HideWindow(Window, #False)
  				
  			ElseIf Event = #PB_Event_Repaint
  				;Debug "repaint"
  				
  			EndIf
  			
  			Event = 0
  			Break
  		EndIf
  	Next
  	
  	; Event not for us ...
  Until Event <> 0

  	
	ProcedureReturn Event
EndProcedure

Procedure.i Hook_WaitWindowEvent(Timeout.i = -1)

	If ListSize(AllocatedDownloadDialogs()) = 0
		If Timeout = -1
			ProcedureReturn WaitWindowEvent()
		Else
			ProcedureReturn  WaitWindowEvent(Timeout)
		EndIf
	EndIf
	
	Protected Event.i, forceReturn.i

	; Calculate end time
	If Timeout <> -1
		Timeout + ElapsedMilliseconds()
	EndIf

	Repeat
	
		; Capture events
		Event = Hook_WindowEvent(@forceReturn)
		If Event <> 0 Or forceReturn
			;Debug Event
			Break
		EndIf
		
		; Delay if no event was found
		Delay(1)
		
	Until (Timeout <> -1 And Timeout < ElapsedMilliseconds())
		
	ProcedureReturn Event
EndProcedure

Macro WindowEvent()
	Hook_WindowEvent()	
EndMacro

Macro WaitWindowEvent(Timeout = -1)
	Hook_WaitWindowEvent(Timeout)
EndMacro
; IDE Options = PureBasic 5.11 (Windows - x64)
; CursorPosition = 602
; FirstLine = 439
; Folding = B8--
; EnableXP